@@ -2,7 +2,7 @@
 #
 # The fisx library for X-Ray Fluorescence
 #
-# Copyright (c) 2014-2016 European Synchrotron Radiation Facility
+# Copyright (c) 2014-2017 European Synchrotron Radiation Facility
 #
 # This file is part of the fisx X-ray developed by V.A. Sole
 #
@@ -198,6 +198,7 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                 // They are ordered by increasing increasing binding energy
                 std::string::size_type iString;
                 std::string ele;
+                std::string lastEle="dummytext";
                 std::string family;
                 std::map<std::string, std::map<std::string, double> > tmpResult;
                 std::map<std::string, std::map<std::string, double> >::const_iterator c_it;
@@ -208,19 +209,37 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                 for (iPeakFamily = 0 ; iPeakFamily < sampleLayerPeakFamilies[iLayer].size(); iPeakFamily++)
                 {
                     iString = sampleLayerPeakFamilies[iLayer][iPeakFamily].first.find(' ');
-                    ele = sampleLayerPeakFamilies[iLayer][iPeakFamily].first.substr(0, iString);
                     family = sampleLayerPeakFamilies[iLayer][iPeakFamily].first.substr(iString + 1, \
                                     sampleLayerPeakFamilies[iLayer][iPeakFamily].first.size() - iString - 1);
-                    // The secondary rates are already corrected for the beam intensity reaching the layer
-                    tmpResult = elementsLibrary.getExcitationFactors(ele, \
+                    ele = sampleLayerPeakFamilies[iLayer][iPeakFamily].first.substr(0, iString);
+                    if (ele != lastEle)
+                    {
+                        // The secondary rates NOT corrected for the beam intensity reaching the layer
+                        // The secondary rates NOT corrected for mass of element fraction in layer
+                        tmpResult = elementsLibrary.getExcitationFactors(ele, \
                                                                      energies[iRay], \
-                                                                     weights[iRay] * sampleLayerWeight[iLayer]);
+                                                                     1.0);
+                        lastEle = ele;
+                    }
                     // and add the energies and rates to the sampleLayerLines
                     for (c_it = tmpResult.begin(); c_it != tmpResult.end(); ++c_it)
                     {
                         // be carefull not to add twice an element
                         if (c_it->first.compare(0, family.length(), family) == 0)
                         {
+                            mapIt2 = c_it->second.find("rate");
+                            if ((mapIt2->second * sampleLayerComposition[ele]) <= 0.0)
+                            {
+                                // std::cout << " Lower equal zero " << mapIt2->second << std::endl;
+                                continue;
+                            }
+                            if (secondaryCalculationLimit > 0.0)
+                            {
+                                if ((mapIt2->second * sampleLayerComposition[ele]) < secondaryCalculationLimit)
+                                {
+                                    continue;
+                                };
+                            }
                             mapIt2 = c_it->second.find("energy");
                             if (mapIt2->second < minimumExcitationEnergy)
                             {
@@ -228,8 +247,10 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                             }
                             sampleLayerEnergies[iLayer].push_back(mapIt2->second);
                             mapIt2 = c_it->second.find("rate");
-                            sampleLayerRates[iLayer].push_back(mapIt2->second *
-                                                               sampleLayerComposition[ele]);
+                            // Store rates already corrected for the beam intensity reaching the layer
+                            // Store rates already corrected for the element mass fraction
+                            sampleLayerRates[iLayer].push_back(mapIt2->second * sampleLayerComposition[ele] * \
+                                                               weights[iRay] * sampleLayerWeight[iLayer]);
                             sampleLayerFamilies[iLayer].push_back(iString);
                             sampleLayerEnergyNames[iLayer].push_back(ele + \
                                                                      " " +\
@@ -435,7 +456,7 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                 // calculate escape ratio assuming normal incidence on detector surface
                                 escapeRates = detector.getEscape(energy, \
                                                                  elementsLibrary, \
-                                                                 c_it->first, \
+                                                                 elementName + c_it->first, \
                                                                  updateEscape);
                                 updateEscape = 0;
                             }
@@ -520,18 +541,13 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                     {
                         if (iLayer == jLayer)
                         {
+                            double criteriumMax=secondaryCalculationLimit;
                             // intralayer secondary
                             for(iLambda = 0; iLambda < sampleLayerEnergies[jLayer].size(); iLambda++)
                             {
                                 // analogous to incident beam
                                 if (energyThreshold > sampleLayerEnergies[jLayer][iLambda])
                                     continue;
-                                // TODO: Make the intensity threshold accessible to the user
-                                if (sampleLayerRates[jLayer][iLambda] < (secondaryCalculationLimit * sampleLayerWeight[iLayer]))
-                                {
-                                    //std::cout << "Skipping due to weight " << std::endl;
-                                    continue;
-                                }
                                 bool calculate;
                                 calculate = true;
                                 if (excitationFactorsCache.find(elementName) != excitationFactorsCache.end())
@@ -552,6 +568,7 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                 }
                                 tmpExcitationFactors = excitationFactorsCache[elementName] \
                                                         [sampleLayerEnergies[jLayer][iLambda]];
+                                double criterium;
                                 for (c_it = result.begin(); c_it != result.end(); ++c_it)
                                 {
                                     if (tmpExcitationFactors.find(c_it->first) == tmpExcitationFactors.end())
@@ -561,6 +578,16 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                     mapIt = result[c_it->first].find("mu_1_i");
                                     if (mapIt == result[c_it->first].end())
                                         throw std::runtime_error(" mu_1_i key. Mass attenuation not present???");
+                                    criterium = tmpExcitationFactors[c_it->first]["rate"] * \
+                                                    sampleLayerRates[jLayer][iLambda];
+                                    criterium /= (primaryExcitationFactors[c_it->first]["rate"] * \
+                                                     sampleLayerWeight[iLayer]);
+                                    if (criterium <= criteriumMax)
+                                    {
+                                        // std::cout << " criterium = " << criterium;
+                                        // std::cout << " criterium Max = " << criteriumMax << std::endl;
+                                        continue;
+                                    }
                                     mu_1_i = mapIt->second;
                                     tmpDouble = Math::deBoerL0(mu_1_lambda / sinAlphaIn,
                                                                mu_1_i / sinAlphaOut,
@@ -592,17 +619,41 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                     result[c_it->first]["secondary"] += tmpDouble;
                                     result[c_it->first]["rate"] += tmpDouble * \
                                                                    result[c_it->first]["efficiency"];
+                                    if (criteriumMax > 0.0)
+                                    {
+                                        if ((tmpDouble/result[c_it->first]["primary"]) < 0.00001)
+                                        {
+                                            if (criterium > criteriumMax)
+                                            {
+                                                criteriumMax = criterium;
+                                            }
+                                        }
+                                    }
                                 }
                             }
                         }
                         else
                         {
+                            if ((sampleLayerWeight[jLayer] / sampleLayerWeight[iLayer]) < 1.0E-4)
+                            {
+                                // The incident reaching the layer originating the secondary excitation
+                                // is relatively much weaker. In the common energy range of XRF interest,
+                                // one will certainly recover less than of 50 % of the incident photons
+                                // reaching the jLayer originating the secondary excitation "beam".
+                                // An additional factor 100 is accounted for because of the possible ratio
+                                // of attenuation coefficient of the element at incident and secondary excitation
+                                // photon energies
+                                continue;
+                            }
+
+                            double criterium;
                             // continue;
                             mu_2_lambda = muTotal[jLayer];
                             density_2 = sampleLayerDensity[jLayer];
                             thickness_2 = sampleLayerThickness[jLayer];
                             if (iLayer < jLayer)
                             {
+                                double criteriumMax = secondaryCalculationLimit;
                                 // interlayer case a)
                                 for(iLambda = 0;
                                     iLambda < sampleLayerEnergies[jLayer].size(); \
@@ -612,12 +663,6 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                     energy = sampleLayerEnergies[jLayer][iLambda];
                                     if (energyThreshold > energy)
                                         continue;
-                                    // TODO: Make the intensity threshold accessible to the user
-                                    if (sampleLayerRates[jLayer][iLambda] < (secondaryCalculationLimit * sampleLayerWeight[iLayer]))
-                                    {
-                                        //std::cout << "Skipping due to weight " << std::endl;
-                                        continue;
-                                    }
                                     bool calculate;
                                     calculate = true;
                                     if (excitationFactorsCache.find(elementName) != excitationFactorsCache.end())
@@ -652,6 +697,13 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                         {
                                             continue;
                                         }
+                                        criterium = tmpExcitationFactors[c_it->first]["rate"] * \
+                                                    sampleLayerRates[jLayer][iLambda];
+                                        criterium /= (primaryExcitationFactors[c_it->first]["rate"] * \
+                                                     sampleLayerWeight[iLayer]);
+                                        if (criterium <= criteriumMax)
+                                            continue;
+
                                         mapIt = result[c_it->first].find("mu_1_i");
                                         if (mapIt == result[c_it->first].end())
                                             throw std::runtime_error(" mu_1_i key. Mass attenuation not present???");
@@ -701,20 +753,36 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                         result[c_it->first]["secondary"] += tmpDouble;
                                         result[c_it->first]["rate"] += tmpDouble * \
                                                                        result[c_it->first]["efficiency"];
+                                        if (criteriumMax > 0.0)
+                                        {
+                                            if ((tmpDouble/result[c_it->first]["primary"]) < 0.00001)
+                                            {
+                                                if (criterium > criteriumMax)
+                                                {
+                                                    criteriumMax = criterium;
+                                                }
+                                            }
+                                        }
                                     }
                                 }
                             }
                             if (iLayer > jLayer)
                             {
+                                double criteriumMax = secondaryCalculationLimit;
                                 // interlayer case b)
                                 double layerFactor;
                                 layerFactor = std::exp(-mu_2_lambda * density_2 * thickness_2/sinAlphaIn);
+                                /*
+                                This test has been removed in favour of the one dealing with the relative
+                                weights of each layer. Indeed, the relative amount of secondary excitation
+                                can be huge important. Other point is if we are going to detect something.
                                 if (layerFactor < 0.001)
                                 {
                                     // No need to calculate anything, the top layer attenuates too much the
                                     // incoming beam
                                     continue;
                                 }
+                                */
                                 for(iLambda = 0;
                                     iLambda < sampleLayerEnergies[jLayer].size(); \
                                     iLambda++)
@@ -723,12 +791,6 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                     energy = sampleLayerEnergies[jLayer][iLambda];
                                     if (energyThreshold > energy)
                                         continue;
-                                    // TODO: Make the intensity threshold accessible to the user
-                                    if (sampleLayerRates[jLayer][iLambda] < (secondaryCalculationLimit * sampleLayerWeight[iLayer]))
-                                    {
-                                        //std::cout << "Skipping due to weight " << std::endl;
-                                        continue;
-                                    }
                                     bool calculate;
                                     calculate = true;
                                     if (excitationFactorsCache.find(elementName) != excitationFactorsCache.end())
@@ -763,6 +825,12 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                         {
                                             continue;
                                         }
+                                        criterium = tmpExcitationFactors[c_it->first]["rate"] * \
+                                                    sampleLayerRates[jLayer][iLambda];
+                                        criterium /= (primaryExcitationFactors[c_it->first]["rate"] * \
+                                                     sampleLayerWeight[iLayer]);
+                                        if (criterium <= criteriumMax)
+                                            continue;
                                         mapIt = result[c_it->first].find("mu_1_i");
                                         if (mapIt == result[c_it->first].end())
                                             throw std::runtime_error(" mu_1_i key. Mass attenuation not present???");
@@ -809,6 +877,16 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                         result[c_it->first]["secondary"] += tmpDouble;
                                         result[c_it->first]["rate"] += tmpDouble * \
                                                                        result[c_it->first]["efficiency"];
+                                        if (criteriumMax > 0.0)
+                                        {
+                                            if ((tmpDouble/result[c_it->first]["primary"]) < 0.00001)
+                                            {
+                                                if (criterium > criteriumMax)
+                                                {
+                                                    criteriumMax = criterium;
+                                                }
+                                            }
+                                        }
                                     }
                                 }
                             }
@@ -824,9 +902,9 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                     if (detector.hasMaterialComposition() || (detector.getMaterialName().size() > 0 ))
                     {
                         // calculate (if needed) escape ratio
-                        escapeRates = detector.getEscape(energy, \
+                        escapeRates = detector.getEscape(result[c_it->first]["energy"], \
                                                          elementsLibrary, \
-                                                         c_it->first, \
+                                                         elementName + c_it->first, \
                                                          updateEscape);
                         if (escapeRates.size())
                         {
@@ -845,6 +923,7 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                     }
                                     actualResult[key][iLayer][tmpString]["energy"] = mapIt->second;
                                     actualResult[key][iLayer][tmpString]["rate"] = 0.0;
+                                    actualResult[key][iLayer][tmpString]["ratio"] = 0.0;
                                     actualResult[key][iLayer][tmpString]["primary"] = 0.0;
                                     actualResult[key][iLayer][tmpString]["secondary"] = 0.0;
                                 }
@@ -855,6 +934,7 @@ std::map<std::string, std::map<int, std::map<std::string, std::map<std::string,
                                 }
                                 totalEscape += mapIt->second;
                                 actualResult[key][iLayer][tmpString]["rate"] += mapIt->second * result[c_it->first]["rate"];
+                                actualResult[key][iLayer][tmpString]["ratio"] = mapIt->second;
                                 // The only meaning of filling "primary" and "secondary" for a escape peak is in order to
                                 // be able to evaluate the ratio without having to refer to the actual parent line.
                                 actualResult[key][iLayer][tmpString]["primary"] += mapIt->second * result[c_it->first]["primary"];
